/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: ftp client for web browser

*/

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>

#include <algorithm>
#include <set.h>
#include <strstream.h>
#include <fstream.h>

#include <config.h>
#include <logf.h>
#include <stringtools.h>
#include <dboxversion.h>
#include <mtoolrfc.h>
#include <htmltools.h>
#include <nls.h>
#include "online.h"
#include "url.h"
#include "htmlview.h"
#include "writemail.h"
#include "basic_www.h"

static const char *skip_e(const char *buffer)
 {
  while(*buffer!=0 && *buffer!=' ' && *buffer!='\r' && *buffer!='\n')
    buffer++;

  while(*buffer==' ')
    buffer++;

  return buffer;
 }

static const char *skip_line(const char *buffer)
 {
  while(*buffer!='\r' && *buffer!='\n')
    buffer++;

  while(*buffer=='\r' || *buffer=='\n')
    buffer++;

  return buffer;
 }

static const char *append_e(const char *buffer, string &data)
 {
  while(*buffer!=0 && *buffer!=' ' && *buffer!='\r' && *buffer!='\n')
   {
    data+=(*buffer);
    buffer++;
   }

  while(*buffer!=0 && *buffer==' ')
    buffer++;

  return buffer;
 }

void do_basic_wwwt::format_ftp_dir(const char *buffer, string &html)
 {
  html+="<table>\n";

  while(buffer[0]!=0)
   {
    if((buffer[0]=='d' && buffer[1]=='r') ||
       (buffer[0]=='l' && buffer[1]=='r') ||
       (buffer[0]=='-' && buffer[1]=='r'))
     {
      string permissions, size, month,
             day_of_month, time_year, file;

      buffer=append_e(buffer, permissions); 
      buffer=skip_e(buffer); // link count
      buffer=skip_e(buffer); // owner
      buffer=skip_e(buffer); // group
      buffer=append_e(buffer, size);
      buffer=append_e(buffer, month);
      buffer=append_e(buffer, day_of_month);
      buffer=append_e(buffer, time_year);
      buffer=append_e(buffer, file);

      if(file!="")
       {
        html+="<tr>";
        html+="<td>"+htmlize_string(permissions)+"</td>"
              "<td>"+month+" "+day_of_month+"</td>"
              "<td>"+time_year+"</td>";
        html+="<td align=right>"+size+"</td>\n";
        html+="<td><a href=\""+htmlize_string(path);
        if(path=="" || path[path.size()-1]!='/') html+='/';
        html+=htmlize_string(file)+"\">"+htmlize_string(file)+"</a></td>";
        html+="</tr>\n";
       }
     }

    buffer=skip_line(buffer);
   }

  html+="</table>\n";
 }

string do_basic_wwwt::get_content_type(const string &filename)
 {
  const char *ext;

  ext=strrchr(filename.c_str(), '.');
  if(ext==NULL) ext="";

  if(stricmp(ext, ".htm")==0 ||
     stricmp(ext, ".html")==0)
    return "text/html";
  else if(stricmp(ext, ".txt")==0 ||
          stricmp(ext, ".lsm")==0 ||
          stricmp(ext, ".tex")==0 ||
          stricmp(ext, ".c")==0 ||
          stricmp(ext, ".h")==0 ||
          stricmp(filename.c_str(), "README")==0 ||
          stricmp(filename.c_str(), "MD5SUMS")==0 ||
          stricmp(filename.c_str(), "COPYING")==0 ||
          stricmp(filename.c_str(), "ChangeLog")==0 ||
          stricmp(filename.c_str(), "TRANS.TBL")==0 ||
          stricmp(filename.c_str(), "INDEX")==0)
    return "text/plain";
  else
    return "application/octet-stream";
 }

bool do_basic_wwwt::get_and_format_ftp()
 {
  ibsockett socket;
  word port_nr;

  port_nr=atol(port.c_str());
  if(port_nr==0) port_nr=21;

  if(socket.open("FTP", host, port_nr))
   {
    string new_url="error://";

    if(socket.get_error_type()==sockett::SOCKET_ERROR)
      new_url+="open.socket/";
    else
      new_url+="open.lookup/";

    new_url+=i2string(socket.get_error_number())+"/"+url;
    parse_url(new_url, protocol, host, port, path);
    return get_and_format_error();
   }

  enum ftp_statet 
   { START, USER, PASS, CWD1, CWD2, TYPE_I, TYPE_A, 
     PASV, RETR_LIST, TRANSFER_DONE, QUIT, DONE }
    ftp_state=START;

  enum ftp_substatet
   { RETR, LIST, LIST_FILE } ftp_substate=RETR;

  string request, session, filename, http, html;
  ibsockett data_socket;
  unsigned long file_size=0;

  while(ftp_state!=DONE)
   {
    request="";

    switch(ftp_state)
     {
     case START:
      break;

     case USER:
      request="USER anonymous\r\n";
      break;

     case PASS:
      request="PASS debris@\r\n";
      break;

     case CWD1:
      if(path=="")
        request="CWD /\r\n";
      else
        request="CWD "+path+"\r\n";
      break;

     case CWD2:
       {
        const char *path_p=path.c_str();
        const char *file=strrchr(path_p, '/');
        if(file!=NULL)
         {
          filename=file+1;
          string new_path;
          for(; path_p<file; path_p++)
            new_path+=path_p[0];
          if(new_path=="") new_path="/";
          request="CWD "+new_path+"\r\n";
         }
        else
          request="CWD "+path+"\r\n";
       }
      break;

     case TYPE_A:
      request="TYPE A\r\n";
      break;

     case TYPE_I:
      request="TYPE I\r\n";
      break;

     case PASV:
      request="PASV\r\n";
      break;

     case TRANSFER_DONE:
      break;

     case RETR_LIST:
      switch(ftp_substate)
       {
       case RETR: request="RETR "+filename+"\r\n"; break;
       case LIST_FILE: request="LIST "+filename+"\r\n"; break;
       case LIST: request="LIST\r\n"; break;
       }
      break;

     case QUIT:
      request="QUIT\r\n";
      break;

     default:
      break;
     }

    if(request!="")
     {
      socket.write(request.c_str(), request.size());
      session+="-&gt; "+htmlize_string(request);
     }

    char line[1000];

    do
     {
      if(socket.gets(line, 999)==NULL)
       {
        string new_url="error://read/"+
                       i2string(socket.get_error_number())+"/"+url;
        parse_url(new_url, protocol, host, port, path);
        return get_and_format_error();
       }
 
      char *tptr;
      tptr=strchr(line, '\r'); if(tptr!=NULL) *tptr=0; 
      tptr=strchr(line, '\n'); if(tptr!=NULL) *tptr=0; 

      session+="&lt;- "+htmlize_string(line)+"\r\n";
     }
    while(!isdigit(line[0]) || line[3]!=' ');

    session+="\r\n";

    unsigned response_code=atoi(line);

    bool error=FALSE;
    switch(ftp_state)
     {
     case START:
      error=(response_code!=220);
      ftp_state=USER;
      session+="** STATE USER\n";
      break;

     case USER:
      if(response_code==230)
       {
        ftp_state=CWD1;
        session+="** STATE CWD1\n";
       }
      else if(response_code==331)
       {
        ftp_state=PASS;
        session+="** STATE PASS\n";
       }
      else
        error=TRUE;
      break;

     case PASS:
      if(response_code==230)
       {
        ftp_state=CWD1;
        session+="** STATE CWD1\n";
       }
      else
        error=TRUE;
      break;

     case TYPE_I:
     case TYPE_A:
      ftp_state=PASV;
      session+="** STATE PASV\n";
      break;

     case CWD1:
      if(response_code==250)
       {
        ftp_state=TYPE_A;
        ftp_substate=LIST;
        session+="** STATE TYPE_A\n";
       }
      else
       {
        ftp_state=CWD2;
        session+="** STATE CWD2\n";
       }
      break;

     case CWD2:
      if(response_code==250)
       {
        ftp_state=TYPE_I;
        ftp_substate=LIST_FILE;
        session+="** STATE TYPE_I\n";
       }
      else
        error=TRUE;
      break;

     case PASV:
      if(response_code==227)
       {
        char *tptr=line+3;
        while(*tptr!=0 && !isdigit(*tptr)) tptr++;
        if(!isdigit(*tptr))
          error=TRUE;
        else
         {
          int A1, A2, A3, A4, a1, a2;
          sscanf(tptr, "%d,%d,%d,%d,%d,%d",
            &A1, &A2, &A3, &A4, &a1, &a2);
          unsigned data_port=(a1<<8)+a2;
          char data_host[TEXTLEN];
          snprintf(data_host, TEXTLEN, "%d.%d.%d.%d", A1, A2, A3, A4);

          if(data_socket.open("FTP-DATA", data_host, data_port))
           {
            string new_url="error://ftp/Failed to open data connection";
            parse_url(new_url, protocol, host, port, path);
            return get_and_format_error(); 
           }

          if(ftp_substate==RETR && file_size!=0)
            data_socket.set_progress_bar(TRUE, file_size);

          ftp_state=RETR_LIST;
          session+="** STATE RETR_LIST\n";
         }        
       }
      else
        error=TRUE;
      break;

     case RETR_LIST:
      if(response_code==150 ||
         response_code==125)
       {
         {
          char *buffer=(char *)malloc(MAX_SIZE);
          if(buffer==(char *)NULL) 
            logf.internalerror("get_and_format_ftp", "out of memory");
          size_t size=data_socket.read(buffer, MAX_SIZE);

          switch(ftp_substate)
           {
           case RETR:
            http=
              "HTTP/1.0 200 Ok\r\n"
              "Content-Type: "+get_content_type(filename)+"\r\n"
              "\r\n";

            http.append(buffer, size);
            break;

           case LIST:
            source.assign(buffer, size);

            if(size==MAX_SIZE)
              buffer[MAX_SIZE]=0;
            else
              buffer[size+1]=0;

            format_ftp_dir(buffer, html);
            break;

           case LIST_FILE:
            if(size==MAX_SIZE)
              buffer[MAX_SIZE]=0;
            else
              buffer[size+1]=0;

             {
              string size;
              const char *tptr=buffer;
              tptr=skip_e(tptr); // permissions
              tptr=skip_e(tptr); // link count
              tptr=skip_e(tptr); // owner
              tptr=skip_e(tptr); // group
              tptr=append_e(tptr, size);
              file_size=atol(size.c_str());
             }
            break;
           }

          free(buffer);
         }

        data_socket.close();
        socket.draw_box();
        ftp_state=TRANSFER_DONE;
        session+="** STATE TRANSFER_DONE\n";
       }
      else
        error=TRUE;
      break;

     case TRANSFER_DONE:
      if(ftp_substate==LIST_FILE)
       {
        ftp_substate=RETR;
        ftp_state=PASV;
        session+="** STATE PASV\n";
       }
      else
       {
        ftp_state=QUIT;
        session+="** STATE QUIT\n";
       }
      break;

     case QUIT:
      ftp_state=DONE;
      session+="** STATE DONE\n";
      break;

     default:
      error=TRUE;
     }

    if(error)
     {
      string page, title, error;

      title=nls["online.ftperror.title"];
      error=nls["online.ftperror.text"];

      source="<html><head><title>";
      source+=nls["online.error"];
      source+=": ";
      source+=title;
      source+="</title></head>\n";
      source+="<body>\n";
      source+=nls["online.httperror.text"];
      source+="<p>\n<blockquote>\n";
      source+=error;
      source+="</blockquote></p>\n";
      source+="<hr><pre>\n";
      source+=session;
      source+="</pre></body></html>";
      
      return format_html(source);
     }
   }

  socket.close();

  if(http!="")
    return format_http(http);

  if(html!="")
    return format_html(html);

  return TRUE;
 }

